"""Taxprofiler related constants."""


from cg.constants.nextflow import NFX_SAMPLE_HEADER, NFX_SAMPLESHEET_READS_HEADERS
from typing import List

TAXPROFILER_RUN_ACCESSION: str = "run_accession"
TAXPROFILER_INSTRUMENT_PLATFORM: str = "instrument_platform"
TAXPROFILER_FASTA_HEADER: str = "fasta"
TAXPROFILER_SAMPLE_SHEET_HEADERS: List[str] = [
    NFX_SAMPLE_HEADER,
    TAXPROFILER_RUN_ACCESSION,
    TAXPROFILER_INSTRUMENT_PLATFORM,
]

TAXPROFILER_SAMPLE_SHEET_HEADERS += NFX_SAMPLESHEET_READS_HEADERS
TAXPROFILER_SAMPLE_SHEET_HEADERS.append(TAXPROFILER_FASTA_HEADER)
